/*
 * Copyright (C) 2003 Steve Grubb.
 *
 * This file is part of the env_audit program.
 * Libenv_audit version 2.x: 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * For more information, 
 *
 *   visit http://www.web-insights.net/libenv_audit/index.html
 */

/* 
 * functions that are supported:
 *   int execl(const char *path, const char *arg, ...);
 *   int execlp(const char *file, const char *arg, ...);
 *   int execle(const char *path, const char *arg, ..., char * const envp[]);
 *   int execv(const char *path, char *const argv[]);
 *   int execvp(const char *file, char *const argv[]);
 *   int execve(const char *filename, char *const argv[], char *const envp[]); 
 *   FILE *popen(const char *command, const char *type);
 *   int system(const char *string);
 *
 */

#include <stdio.h>		/* defines stdin */
#include <unistd.h>		/* readlink */
#include <stdlib.h>		/* free() */
#include <sys/param.h>		/* MAXPATHLEN */
#include <stdarg.h>             /* defines va_args */
#include <errno.h>		/* defines errno */
#include <syslog.h>		/* for syslog() */
#include <sys/wait.h>
#include <sys/syscall.h>
#include <sys/stat.h>		/* For some file mode flags */
#include <fcntl.h>		/* For some file creation flags */
#define __USE_GNU 1		/* defines RTLD_NEXT */
#include <string.h>		/* strnlen, strcmp */
#include <dlfcn.h>		/* defines dlsym() */
#define VERSION  LIBENV_AUDIT_VERSION

/*
 * -----------------------------------------------------------------
 * ----------------- system library overrides ----------------------
 */
typedef int (*execv_t) (const char *path, char *const argv[]);
typedef int (*execl_t) (const char *path, const char *arg, ...);
typedef int (*execvp_t) (const char *file, char *const argv[]);
typedef int (*execlp_t) (const char *file, const char *arg, ...);
typedef int (*execve_t) (const char *file, char *const argv[],
	char *const envp[]);
typedef int (*execle_t) (const char *path, const char *arg, ...);
typedef FILE *(*popen_t) (const char *command, const char *type);
typedef int (*system_t) (const char *string);


/*
 * 0 = don't do any libenv_audit checking for this process
 * 1 = enable libenv_audit checking for this process
 */
volatile int doing_audit = 0;
const char audit[] = "/usr/bin/env_audit";

/*
 * The function pointers. We switch all calls over to the array variety.
 */
static execv_t real_execv = NULL;
static execvp_t real_execvp = NULL;
static execve_t real_execve = NULL;
static system_t real_system = NULL;

static void *getLibraryFunction(const char *funcName);
static void log_info(const char *func, const char *cmd);

/*
 * -------------- system library implementations -------------------
 */

int execv(const char *path, char *const argv[])
{
    int ret_val;

    if (!real_execv)
	real_execv = (execv_t) getLibraryFunction("execv");
    if (!real_system)
	real_system = (system_t) getLibraryFunction("__libc_system");

    if (doing_audit)
	return real_execv(path, argv);

    doing_audit = 1;
    log_info("execv", path);
    real_system(audit);

    ret_val = real_execv(path, argv);
    doing_audit = 0;
    return ret_val;
}

int execl(const char *path, const char *arg, ...)
{
    int ret_val;
    unsigned int i, num=0;
    va_list ap;
    const char *ptr, **arg_list=NULL;

    if (!real_execv)
	real_execv = (execv_t) getLibraryFunction("execv");
    if (!real_system)
	real_system = (system_t) getLibraryFunction("__libc_system");

    if (arg) {  /* Count the entries */
       num++;
       va_start(ap, arg);
       do {
          ptr = va_arg(ap, char *);
          num++; /* Always incr since we want the final NULL pointer, too */
       } while (ptr);
       va_end(ap);
    }

    if (num) {  /* Make the array */
       arg_list = (const char **)malloc(num * sizeof(char *));
       if (arg_list == NULL) {
          errno = ENOMEM;
          return -1;
       }

       /* Fill the array */
       arg_list[0] = arg;
       va_start(ap, arg);
       for (i=1; i<num; i++)
          arg_list[i] = va_arg(ap, char *);
       va_end(ap);
    }

    if (doing_audit) {
	ret_val =  real_execv(path, arg_list);
        free(arg_list);
        return ret_val;
    }

    doing_audit = 1;
    log_info("execl", path);
    real_system(audit);

    ret_val = real_execv(path, arg_list);
    free(arg_list);
    doing_audit = 0;
    return ret_val;
}

int execvp(const char *file, char *const argv[])
{
    int ret_val;

    if (!real_execvp)
	real_execvp = (execvp_t) getLibraryFunction("execvp");
    if (!real_system)
        real_system = (system_t) getLibraryFunction("__libc_system");

    if (doing_audit)
	return real_execvp(file, argv);

    doing_audit = 1;
    log_info("execvp", file);
    real_system(audit);

    ret_val = real_execvp(file, argv);
    doing_audit = 0;
    return ret_val;
}

int execlp(const char *file, const char *arg, ...)
{
    int ret_val;
    unsigned int i, num=0;
    va_list ap;
    const char *ptr, **arg_list=NULL;

    if (!real_execvp)
	real_execvp = (execvp_t) getLibraryFunction("execvp");
    if (!real_system)
        real_system = (system_t) getLibraryFunction("__libc_system");

    if (arg) { /* Count the entries */
       num++;
       va_start(ap, arg);
       do {
          ptr = va_arg(ap, char *);
          num++; /* Always incr since we want the final NULL pointer, too */
       } while (ptr);
       va_end(ap);
    }

    if (num) { /* Make the array */
       arg_list = (char **)malloc(num * sizeof(char *));
       if (arg_list == NULL) {
          errno = ENOMEM;
          return -1;
       }

       /* Fill the array */
       arg_list[0] = arg;
       va_start(ap, arg);
       for (i=1; i<num; i++)
          arg_list[i] = va_arg(ap, char *);
       va_end(ap);
    }

    if (doing_audit) {
	ret_val =  real_execvp(file, arg_list);
        free(arg_list);
        return ret_val;
    }

    doing_audit = 1;
    log_info("execlp", file);
    real_system(audit);

    ret_val = real_execvp(file, arg_list);
    free(arg_list);
    doing_audit = 0;
    return ret_val;
}

int execve(const char *filename, char *const argv[], char *const envp[])
{
    int ret_val;
    pid_t pid;

    if (!real_execve)
	real_execve = (execve_t) getLibraryFunction("execve");

    if (filename == NULL) {
	errno = EINVAL;
	return -1;
    }

    if (doing_audit || strcmp(filename, audit) == 0)
	return real_execve(filename, argv, envp);

    doing_audit = 1;
    
    log_info("execve", filename);
    pid = fork();
    switch (pid)
    {
        case 0:
	    syscall(__NR_execve, audit, argv, envp);
	    break;
	case -1:
	    break;
	default:
	    waitpid(pid, NULL, 0);
	    break;
    }	

    ret_val = real_execve(filename, argv, envp);
    doing_audit = 0;
    return ret_val;
}

int execle(const char *path, const char *arg, ...)
{
    int ret_val;
    unsigned int i, num_arg=0, num_env=0;
    va_list ap;
    const char *ptr, **arg_list=NULL, **env_list=NULL;

    if (!real_execve)
	real_execve = (execve_t) getLibraryFunction("execve");
    if (!real_system)
	real_system = (system_t) getLibraryFunction("__libc_system");

    if (arg) {
       /* Count the entries */
       num_arg++;
       va_start(ap, arg);
       do {
          ptr = va_arg(ap, char *);
          num_arg++; /* Always incr since we want final NULL pointer, too */
       } while (ptr);
       do {
          ptr = va_arg(ap, char *);
          num_env++; /* Always incr since we want final NULL pointer, too */
       } while (ptr);
       va_end(ap);
    }

    if (num_arg) {
       /* Make the array */
       arg_list = (char **)malloc(num_arg * sizeof(char *));
       if (arg_list == NULL) {
          errno = ENOMEM;
          return -1;
       }
    }
    if (num_env) {
       env_list = (char **)malloc(num_env * sizeof(char *));
       if (env_list == NULL) {
          errno = ENOMEM;
          free(arg_list);
          return -1;
       }
    }

    /* Fill the array */
    if (num_arg) {
       arg_list[0] = arg;
       va_start(ap, arg);
       for (i=1; i<num_arg; i++)
          arg_list[i] = va_arg(ap, char *);
       for (i=0; i<num_env; i++)
          env_list[i] = va_arg(ap, char *);
       va_end(ap);
    }
    if (doing_audit) {
	ret_val =  real_execve(path, arg_list, env_list);
        free(arg_list);
        free(env_list);
        return ret_val;
    }

    doing_audit = 1;
    log_info("execle", path);
    real_system(audit);

    ret_val = real_execve(path, arg_list, env_list);
    free(arg_list);
    free(env_list);
    doing_audit = 0;
    return ret_val;
}

FILE *popen(const char *command, const char *type)
{
    static popen_t real_popen = NULL;

    if (!real_popen)
	real_popen = (popen_t) getLibraryFunction("popen");
    if (!real_system)
	real_system = (system_t) getLibraryFunction("__libc_system");

    if (doing_audit)
	return real_popen(command, type);

    doing_audit = 1;
    log_info("popen", command);
    real_system(audit);
    doing_audit = 0;

    return real_popen(command, type);
}

int system(const char *string)
{
    int ret_val;

    if (!real_system)
	real_system = (system_t) getLibraryFunction("__libc_system");

    if (doing_audit)
	return real_system(string);

    doing_audit = 1;
    log_info("system", string);
    real_system(audit);

    ret_val = real_system(string);
    doing_audit = 0;
    return ret_val;
}

/*
 * returns a pointer to the implementation of 'funcName' in
 * the libc library.  If not found, terminates the program.
 */
static void *getLibraryFunction(const char *funcName)
{
    void *res;

    if ((res = dlsym(RTLD_NEXT, funcName)) == NULL) {
	syslog(LOG_CRIT, "dlsym %s error:%s\n", funcName, dlerror());
	_exit(1);
    }
    doing_audit = 0;
    return res;
}

static char *get_exename(char *name, int size, int pid)
{
    int res;
    char path[4096];

    snprintf(path, 4096, "/proc/%d/exe", pid);
    if ((res = readlink(path, name, size - 1)) == -1)
        name[0] = 0;
    else
        name[res] = 0;
    if (strcmp(name, "/bin/bash") == 0) {
        int f;
        snprintf(path, 4096, "/proc/%d/cmdline", pid);
        f = open(path, O_RDONLY);
        if ( f >= 0 ) {
            int len, i;

            len = read(f, path, size);
            if (len > 0) {
                char **argv, *ptr;
                int cnt=0;

                for (i=0; i<len; i++) {
                    if (path[i] == 0)
                     cnt++;
                }

/*
 * Algorithm - take string and break into array of strings (argv).
 * Scan for just bash or just sh as $0. If they are found, look at 
 * the rest of the arguments to see if a -c is given. If so, the 
 * command follows. 
 *
 * If bash or sh are by themselves, it is the shell. Otherwise, $1 
 * is the command when /bin/bash is reported by readlink of self/exe.
 */

                if ( cnt )
                    argv = (char **)malloc(sizeof(char *) * cnt );
                else { /* Can't do anything */
                    strcpy(name, "/bin/bash");
                    return name;
                }

                /* Create the array */
                ptr = path;
                name[0] = 0;
                for (i=0; i<cnt; i++) {
                    int len = strlen(ptr);
                    argv[i] = (char *)malloc(len+1);
                    strcpy(argv[i], ptr);
                    ptr += len+1;
                }

                /* see if command by itself */
                if ((strcmp(argv[0], "bash") == 0) ||
                    (strcmp(argv[0], "sh") == 0)) {
                    for (i=1; i<cnt; i++) {
                        if ((strcmp(argv[i], "-c") == 0) && (i+1 < cnt)) {
                            int new_len, x;
                            strncpy(name, argv[i+1], size-1);
                            name[size-1] = 0;
                            for(x=0; x<len; x++)
                                if (name[x] == ' ') { name[x] = 0; break; }
                            break;
                        }
                    }
                    if (name[0] == 0) { /* Nothing found, must be from stdin */
                        strcpy(name, "/bin/bash");
                    }
                }
                else if (strcmp(argv[0], "/bin/bash") == 0) { 
                    /* Scan for '-c' */
                    for (i=0; i<cnt; i++) {
                        if ((strcmp(argv[i], "-c") == 0) && (i+1 < cnt)) {
                            int new_len, x;
                            strncpy(name, argv[i+1], size-1);
                            name[size-1] = 0;
                            new_len = strlen(name);
                            for(x=0; x<len; x++)
                                if (name[x] == ' ') { name[x] = 0; break; }
                            break;
                        }
                    }
                    /* If nothing found, argument $1 must be the command */
                    if ((name[0] == 0) && (cnt >= 2)) {
                        strncpy(name, argv[1], size-1);
                        name[size-1] = 0;
                    }
                    if (name[0] == 0) /* Nothing found at all,use what we had */
                        strcpy(name, "/bin/bash");
                }
                else { /* First item isn't shell, so use it */
                        strncpy(name, argv[0], size-1);
                        name[size-1] = 0;
                }
                for (i=0; i<cnt; i++)
                    if (argv[i]) free(argv[i]);
                free(argv);
            }
            close(f);
        }
    }
    return (name);
}

/*
 * This is called to dump the stack to see the call chain to leading up
 * to the event we are interested in.
 */
#define LOGFILE "/tmp/env_trace"
static void log_info(const char *func, const char *cmd)
{
    char exename[MAXPATHLEN];
    char parname[MAXPATHLEN];
    int index;
    int x=0;
    FILE *f=NULL;

    get_exename(exename, MAXPATHLEN, getpid());
    get_exename(parname, MAXPATHLEN, getppid());

    /*
     * Iterate though different numbers to find the next available filename.
     */
    for (x=0; x<10000; x++)
    {
       int j;
       char log_file[PATH_MAX+1];

       snprintf(log_file, sizeof(log_file), "%s%04d.log", LOGFILE, x);
       j = open(log_file, O_CREAT|O_EXCL|O_WRONLY, S_IRUSR|S_IWUSR);
       if (j >= 0)
       {
          f = fdopen(j, "wt");
          if (f)
             break;
          else
             close(j);
       }
    }
    if (f == NULL)
       return ;

    fprintf(f, "Libenv_audit version: %s\n", VERSION);
    fprintf(f, "This is the data gathered in the parent, before exec,\n");
    fprintf(f, "system, or popen was called.\n");
    fprintf(f, "uid=%d\neuid=%d\n",
                getuid(), geteuid());
    fprintf(f, "ppid=%d\n", getppid());
    fprintf(f, "Parent name: %s\n", parname);
    fprintf(f, "pid=%d\n", getpid());
    fprintf(f, "Currently running: %s\n", exename);
    fprintf(f, "Function called: %s\n", func);
    fprintf(f, "Executable is: %s\n", cmd);

    fclose(f);
}

/*
 * -----------------------------------------------------------------
 * ------------- initializer and finalizer--------------------------
 */
static void _intercept_init() __attribute__ ((constructor));
static void _intercept_fini() __attribute__ ((destructor));

static void _intercept_init(void)
{
}

static void _intercept_fini(void)
{
}

